----------------------------------------------------------------------------------------------------
-- FreeFloatComponentJoints
----------------------------------------------------------------------------------------------------
-- Purpose: allow components to behave in free float mode.
--
-- Based on: The unlock component joint script from rafftnix.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class FreeFloatComponentJoints
FreeFloatComponentJoints = {}
FreeFloatComponentJoints.MOD_NAME = g_currentModName

function FreeFloatComponentJoints.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Cylindered, specializations)
end

function FreeFloatComponentJoints.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "loadFreeFloatComponentJointFromXML", FreeFloatComponentJoints.loadFreeFloatComponentJointFromXML)
    SpecializationUtil.registerFunction(vehicleType, "canFreeFloat", FreeFloatComponentJoints.canFreeFloat)
    SpecializationUtil.registerFunction(vehicleType, "setFreeFloatState", FreeFloatComponentJoints.setFreeFloatState)
end

function FreeFloatComponentJoints.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", FreeFloatComponentJoints)
    SpecializationUtil.registerEventListener(vehicleType, "onReadStream", FreeFloatComponentJoints)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteStream", FreeFloatComponentJoints)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", FreeFloatComponentJoints)
    SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", FreeFloatComponentJoints)
end

function FreeFloatComponentJoints:onLoad(savegame)
    self.spec_freeFloatComponentJoints = self[("spec_%s.freeFloatComponentJoints"):format(FreeFloatComponentJoints.MOD_NAME)]
    local spec = self.spec_freeFloatComponentJoints

    spec.isFreeFloating = false
    spec.isDirty = false
    spec.freeFloatComponentJoints = {}

    local i = 0
    while true do
        local key = ("vehicle.freeFloatComponentJoints.freeFloatComponentJoint(%d)"):format(i)

        if not hasXMLProperty(self.xmlFile, key) then
            break
        end

        local entry = {}
        if self:loadFreeFloatComponentJointFromXML(self.xmlFile, key, entry) then
            table.insert(spec.freeFloatComponentJoints, entry)
        end

        i = i + 1
    end
end

function FreeFloatComponentJoints:onReadStream(streamId, connection)
    local isFreeFloating = streamReadBool(streamId)
    self:setFreeFloatState(isFreeFloating, true)
end

function FreeFloatComponentJoints:onWriteStream(streamId, connection)
    local spec = self.spec_freeFloatComponentJoints
    streamWriteBool(streamId, spec.isFreeFloating)
end

function FreeFloatComponentJoints:onUpdateTick(dt)
    local spec = self.spec_freeFloatComponentJoints

    if not self.isServer then
        return
    end

    if spec.isDirty then
        local didUpdate = false
        for _, control in ipairs(spec.freeFloatComponentJoints) do
            if control.targetMinRotation ~= control.lastMinRotation
                or control.targetMaxRotation ~= control.lastMaxRotation then
                local componentJointDesc = self.componentJoints[control.componentJointIndex]
                local movingTool = self:getMovingToolByNode(control.movingToolNode)

                control.alpha = math.min(control.alpha + dt / control.freeFloatDuration, 1)
                control.lastMinRotation = MathUtil.lerp(control.lastMinRotation, control.targetMinRotation, control.alpha)
                control.lastMaxRotation = MathUtil.lerp(control.lastMaxRotation, control.targetMaxRotation, control.alpha)

                self:setComponentJointRotLimit(componentJointDesc, movingTool.rotationAxis, -control.lastMinRotation, control.lastMaxRotation)
                self:setMovingToolDirty(movingTool)
                didUpdate = true
            end
        end

        if not didUpdate then
            spec.isDirty = false
        end
    end
end

function FreeFloatComponentJoints:loadFreeFloatComponentJointFromXML(xmlFile, key, control)
    control.componentJointIndex = Utils.getNoNil(getXMLInt(xmlFile, key .. "#componentJointIndex"), 0)
    control.movingToolNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. "#movingToolNode"), self.i3dMappings)

    if control.movingToolNode == nil then
        g_logManager:xmlError(self.configFileName, ("Moving tool node not set for the free float component joint: %d"):format(control.componentJointIndex))
        return false
    end

    local movingTool = self:getMovingToolByNode(control.movingToolNode)
    -- Backup values from the moving tool.
    control.rotSpeed = movingTool.rotSpeed
    control.freeFloatDuration = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#duration"), 5000)

    control.lastMinRotation = 0
    control.lastMaxRotation = 0
    control.alpha = 0
    control.isAdaptive = Utils.getNoNil(getXMLBool(xmlFile, key .. "#isAdaptive"), true)

    return true
end

function FreeFloatComponentJoints:canFreeFloat()
    return true
end

function FreeFloatComponentJoints:setFreeFloatState(isFreeFloating, noEventSend)
    local spec = self.spec_freeFloatComponentJoints

    if spec.isFreeFloating ~= isFreeFloating then
        FreeFloatComponentJointsEvent.sendEvent(self, isFreeFloating, noEventSend)

        spec.isFreeFloating = isFreeFloating

        local actionEvent = spec.actionEvents[InputAction.RE_TOGGLE_FLOAT]
        if actionEvent ~= nil then
            local text = g_i18n:getText("action_enableFreeFloat"):format(self.typeDesc)
            if not spec.isFreeFloating then
                text = g_i18n:getText("action_disableFreeFloat"):format(self.typeDesc)
            end

            g_inputBinding:setActionEventText(actionEvent.actionEventId, text)
        end

        if self.isServer then
            for _, control in ipairs(spec.freeFloatComponentJoints) do
                local movingTool = self:getMovingToolByNode(control.movingToolNode)
                local targetMinRotation = 0
                local targetMaxRotation = 0

                control.alpha = 0
                if isFreeFloating then
                    movingTool.rotSpeed = 0

                    local rotation = { unpack(movingTool.curRot) }
                    local curRot = rotation[movingTool.rotationAxis]
                    local rotLimit = (movingTool.rotMax - movingTool.rotMin) * 0.5
                    local minRot = rotLimit + curRot
                    local maxRot = rotLimit - curRot

                    if not control.isAdaptive then
                        local curRotAbs = math.abs(curRot)
                        if curRot < 0 then
                            maxRot = curRotAbs + movingTool.rotMax
                        else
                            maxRot = movingTool.rotMax - curRotAbs
                        end

                        rotLimit = movingTool.rotMax - movingTool.rotMin
                        minRot = rotLimit - maxRot
                    end

                    targetMinRotation = minRot
                    targetMaxRotation = maxRot
                else
                    movingTool.rotSpeed = control.rotSpeed
                    local x, y, z = getRotation(movingTool.node)
                    movingTool.curRot = { x, y, z }
                end

                control.targetMinRotation = targetMinRotation
                control.targetMaxRotation = targetMaxRotation
                spec.isDirty = true
            end
        end
    end
end

--- Action events
function FreeFloatComponentJoints.actionEventToggleFloat(self)
    local spec = self.spec_freeFloatComponentJoints

    if self:canFreeFloat() then
        self:setFreeFloatState(not spec.isFreeFloating)
    end
end

function FreeFloatComponentJoints:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
    if self.isClient then
        local spec = self.spec_freeFloatComponentJoints

        self:clearActionEventsTable(spec.actionEvents)

        if isActiveForInput then
            local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.RE_TOGGLE_FLOAT, self, FreeFloatComponentJoints.actionEventToggleFloat, false, true, false, true, nil, nil, true)
            local text = g_i18n:getText("action_enableFreeFloat"):format(self.typeDesc)
            if not spec.isFreeFloating then
                text = g_i18n:getText("action_disableFreeFloat"):format(self.typeDesc)
            end
            g_inputBinding:setActionEventText(actionEventId, text)
            g_inputBinding:setActionEventTextVisibility(actionEventId, true)
            g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_HIGH)
        end

    end
end
